@@ -1,5 +1,6 @@
+#![allow(dead_code)]
 use fastnbt::error::Result;
-use fastnbt::{de::from_bytes, Value};
+use fastnbt::{from_bytes, Value};
 use flate2::read::GzDecoder;
 use serde::Deserialize;
 use std::io::Read;
