@@ -243,41 +243,12 @@ use std::ops::Range;
 
 use crate::de_arrays::ArrayWrapperAccess;
 use crate::error::{Error, Result};
-use crate::Tag;
+use crate::{DeOpts, Tag};
 use byteorder::{BigEndian, ReadBytesExt};
 
+use serde::de::Unexpected;
 use serde::{de, forward_to_deserialize_any};
 
-/// Deserialize into a `T` from some NBT data. See the [`de`] module for more
-/// information.
-///
-/// ```no_run
-/// # use fastnbt::Value;
-/// # use flate2::read::GzDecoder;
-/// # use std::io;
-/// # use std::io::Read;
-/// # use fastnbt::error::Result;
-/// # fn main() -> Result<()> {
-/// # let some_reader = io::stdin();
-/// let mut decoder = GzDecoder::new(some_reader);
-/// let mut buf = vec![];
-/// decoder.read_to_end(&mut buf).unwrap();
-///
-/// let val: Value = fastnbt::de::from_bytes(buf.as_slice())?;
-/// # Ok(())
-/// # }
-/// ```
-///
-/// [`de`]: ./index.html
-pub fn from_bytes<'a, T>(input: &'a [u8]) -> Result<T>
-where
-    T: de::Deserialize<'a>,
-{
-    let mut des = Deserializer::from_bytes(input);
-    let t = T::deserialize(&mut des)?;
-    Ok(t)
-}
-
 /// Deserializer for NBT data. See the [`de`] module for more information.
 ///
 /// [`de`]: ./index.html
@@ -285,6 +256,7 @@ pub struct Deserializer<'de> {
     pub(crate) input: InputHelper<'de>,
     layers: Vec<Layer>,
     last_hint: Option<&'static str>,
+    pub(crate) opts: DeOpts,
 }
 
 impl<'de> Deserializer<'de> {
@@ -292,11 +264,12 @@ impl<'de> Deserializer<'de> {
     /// for more information.
     ///
     /// [`de`]: ./index.html
-    pub fn from_bytes(input: &'de [u8]) -> Self {
+    pub fn from_bytes(input: &'de [u8], opts: DeOpts) -> Self {
         Self {
             input: InputHelper(input),
             layers: vec![],
             last_hint: None,
+            opts,
         }
     }
 }
@@ -375,6 +348,13 @@ where
                 ));
             }
 
+            if size as usize >= de.opts.max_seq_len {
+                return Err(Error::bespoke(format!(
+                    "size ({}) greater than max sequence length ({})",
+                    size, de.opts.max_seq_len,
+                )));
+            }
+
             de.layers.push(Layer::List {
                 remaining_elements: size,
                 element_tag,
@@ -382,15 +362,52 @@ where
 
             visitor.visit_seq(ListAccess::new(de, size))
         }
-        Tag::ByteArray | Tag::IntArray | Tag::LongArray => {
+        Tag::ByteArray => {
             if last_hint == Some("seq") {
                 return Err(Error::bespoke(
                     "expected NBT Array, found seq: use ByteArray, IntArray or LongArray types"
                         .into(),
                 ));
             }
+
             let size = de.input.consume_list_size()?;
-            visitor.visit_map(ArrayWrapperAccess::new(de, size, tag))
+
+            // visitor.visit_borrowed_bytes(bs)
+            visitor.visit_map(ArrayWrapperAccess::bytes(
+                de,
+                size.try_into()
+                    .map_err(|_| Error::bespoke("nbt array size was negative".to_string()))?,
+            )?)
+        }
+        Tag::IntArray => {
+            if last_hint == Some("seq") {
+                return Err(Error::bespoke(
+                    "expected NBT Array, found seq: use ByteArray, IntArray or LongArray types"
+                        .into(),
+                ));
+            }
+
+            let size = de.input.consume_list_size()?;
+            visitor.visit_map(ArrayWrapperAccess::ints(
+                de,
+                size.try_into()
+                    .map_err(|_| Error::bespoke("nbt array size was negative".to_string()))?,
+            )?)
+        }
+        Tag::LongArray => {
+            if last_hint == Some("seq") {
+                return Err(Error::bespoke(
+                    "expected NBT Array, found seq: use ByteArray, IntArray or LongArray types"
+                        .into(),
+                ));
+            }
+
+            let size = de.input.consume_list_size()?;
+            visitor.visit_map(ArrayWrapperAccess::longs(
+                de,
+                size.try_into()
+                    .map_err(|_| Error::bespoke("nbt array size was negative".to_string()))?,
+            )?)
         }
         // This would really only occur when we encounter a list where the
         // element type is 'End', but we specifically handle that case, so we
@@ -430,9 +447,21 @@ impl<'de> InputHelper<'de> {
         Ok(s)
     }
 
-    fn consume_bytes_unchecked(&mut self, size: i32) -> Result<&'de [u8]> {
+    fn consume_size_prefixed_bytes(&mut self) -> Result<&'de [u8]> {
+        let len = self.0.read_u16::<BigEndian>()? as usize;
+        let str_data = self.subslice(0..len)?;
+
+        self.0 = &self.0[len..];
+        Ok(str_data)
+    }
+
+    pub(crate) fn consume_bytes(&mut self, size: i32) -> Result<&'de [u8]> {
         let size: usize = size.try_into().map_err(|_| Error::invalid_size(size))?;
-        let bs = &self.0[..size];
+        self.consume_bytes_usize(size)
+    }
+
+    pub(crate) fn consume_bytes_usize(&mut self, size: usize) -> Result<&'de [u8]> {
+        let bs = self.subslice(0..size)?;
         self.0 = &self.0[size..];
         Ok(bs)
     }
@@ -474,15 +503,15 @@ impl<'de> InputHelper<'de> {
             }
             Tag::ByteArray => {
                 let size = self.consume_list_size()?;
-                self.consume_bytes_unchecked(size)?;
+                self.consume_bytes(size)?;
             }
             Tag::IntArray => {
                 let size = self.consume_list_size()?;
-                self.consume_bytes_unchecked(size * 4)?;
+                self.consume_bytes_usize(try_size(size, 4)?)?;
             }
             Tag::LongArray => {
                 let size = self.consume_list_size()?;
-                self.consume_bytes_unchecked(size * 8)?;
+                self.consume_bytes_usize(try_size(size, 8)?)?;
             }
             Tag::Compound => {
                 // Need to loop and ignore each value until we reach an end tag.
@@ -524,7 +553,7 @@ impl<'de> InputHelper<'de> {
 impl<'de, 'a> de::Deserializer<'de> for &'a mut Deserializer<'de> {
     type Error = Error;
 
-    forward_to_deserialize_any!(struct map identifier i8 i16 i32 i64 u8 u16 u32 u64 f32 f64 str string tuple);
+    forward_to_deserialize_any!(struct map identifier char i8 i16 i32 i64 u8 u16 u32 u64 f32 f64 str string tuple);
 
     fn is_human_readable(&self) -> bool {
         false
@@ -613,14 +642,6 @@ impl<'de, 'a> de::Deserializer<'de> for &'a mut Deserializer<'de> {
         }
     }
 
-    #[inline]
-    fn deserialize_char<V>(self, _visitor: V) -> Result<V::Value>
-    where
-        V: de::Visitor<'de>,
-    {
-        unimplemented!("char")
-    }
-
     #[inline]
     fn deserialize_bytes<V>(self, visitor: V) -> Result<V::Value>
     where
@@ -653,19 +674,19 @@ impl<'de, 'a> de::Deserializer<'de> for &'a mut Deserializer<'de> {
 
                 match el {
                     Tag::Byte => {
-                        let bs = self.input.consume_bytes_unchecked(size)?;
+                        let bs = self.input.consume_bytes(size)?;
                         visitor.visit_borrowed_bytes(bs)
                     }
                     Tag::Short => {
-                        let bs = self.input.consume_bytes_unchecked(size * 2)?;
+                        let bs = self.input.consume_bytes_usize(try_size(size, 2)?)?;
                         visitor.visit_borrowed_bytes(bs)
                     }
                     Tag::Int => {
-                        let bs = self.input.consume_bytes_unchecked(size * 4)?;
+                        let bs = self.input.consume_bytes_usize(try_size(size, 4)?)?;
                         visitor.visit_borrowed_bytes(bs)
                     }
                     Tag::Long => {
-                        let bs = self.input.consume_bytes_unchecked(size * 8)?;
+                        let bs = self.input.consume_bytes_usize(try_size(size, 8)?)?;
                         visitor.visit_borrowed_bytes(bs)
                     }
                     _ => Err(Error::bespoke(format!(
@@ -680,32 +701,35 @@ impl<'de, 'a> de::Deserializer<'de> for &'a mut Deserializer<'de> {
             } => match tag {
                 Tag::ByteArray => {
                     let size = self.input.consume_list_size()?;
-                    let bs = self.input.consume_bytes_unchecked(size)?;
+                    let bs = self.input.consume_bytes(size)?;
                     visitor.visit_borrowed_bytes(bs)
                 }
                 Tag::IntArray => {
                     let size = self.input.consume_list_size()?;
-                    let bs = self.input.consume_bytes_unchecked(size * 4i32)?;
+                    let bs = self.input.consume_bytes_usize(try_size(size, 4)?)?;
                     visitor.visit_borrowed_bytes(bs)
                 }
                 // This allows us to borrow blockstates rather than copy them.
                 Tag::LongArray => {
                     let size = self.input.consume_list_size()?;
-                    let bs = self.input.consume_bytes_unchecked(size * 8i32)?;
+                    let bs = self.input.consume_bytes_usize(try_size(size, 8)?)?;
                     visitor.visit_borrowed_bytes(bs)
                 }
+                Tag::String => {
+                    let s = self.input.consume_size_prefixed_bytes()?;
+                    visitor.visit_borrowed_bytes(s)
+                }
                 _ => Err(Error::bespoke(format!("expected bytes, found {:?}", tag))),
             },
         }
     }
 
     #[inline]
-    fn deserialize_byte_buf<V>(self, _visitor: V) -> Result<V::Value>
+    fn deserialize_byte_buf<V>(self, visitor: V) -> Result<V::Value>
     where
         V: de::Visitor<'de>,
     {
-        // How do we even get here? Vec and slices don't call this.
-        unimplemented!("byte_buf")
+        self.deserialize_bytes(visitor)
     }
 
     #[inline]
@@ -736,19 +760,42 @@ impl<'de, 'a> de::Deserializer<'de> for &'a mut Deserializer<'de> {
     }
 
     #[inline]
-    fn deserialize_unit_struct<V>(self, _name: &'static str, _visitor: V) -> Result<V::Value>
+    fn deserialize_unit_struct<V>(self, _name: &'static str, visitor: V) -> Result<V::Value>
     where
         V: de::Visitor<'de>,
     {
-        todo!("unit_struct")
+        self.deserialize_unit(visitor)
     }
 
     #[inline]
-    fn deserialize_newtype_struct<V>(self, _name: &'static str, visitor: V) -> Result<V::Value>
+    fn deserialize_newtype_struct<V>(self, name: &'static str, visitor: V) -> Result<V::Value>
     where
         V: de::Visitor<'de>,
     {
-        visitor.visit_newtype_struct(self)
+        let target_tag = match name {
+            crate::BYTE_ARRAY_TOKEN => Tag::ByteArray,
+            crate::INT_ARRAY_TOKEN => Tag::IntArray,
+            crate::LONG_ARRAY_TOKEN => Tag::LongArray,
+            _ => return visitor.visit_newtype_struct(self),
+        };
+
+        let data_tag = *match self.layers.last() {
+            Some(Layer::Compound { current_tag, .. }) => current_tag
+                .as_ref()
+                .ok_or_else(|| Error::bespoke("deserialize: did not know value's tag".to_string())),
+            Some(Layer::List { element_tag, .. }) => Ok(element_tag),
+            None => Err(Error::bespoke(
+                "deserialize: not in compound or list".to_string(),
+            )),
+        }?;
+
+        if target_tag == data_tag {
+            consume_value(self, visitor, target_tag)
+        } else {
+            Err(Error::bespoke(format!(
+                "expected {data_tag}, found {target_tag}"
+            )))
+        }
     }
 
     #[inline]
@@ -756,12 +803,12 @@ impl<'de, 'a> de::Deserializer<'de> for &'a mut Deserializer<'de> {
         self,
         _name: &'static str,
         _len: usize,
-        _visitor: V,
+        visitor: V,
     ) -> Result<V::Value>
     where
         V: de::Visitor<'de>,
     {
-        todo!("tuple_struct")
+        self.deserialize_seq(visitor)
     }
 
     #[inline]
@@ -800,12 +847,12 @@ impl<'de, 'a> de::Deserializer<'de> for &'a mut Deserializer<'de> {
             Layer::Compound {
                 current_tag: _,
                 stage: _,
-            } => todo!("compound(none)"), // ???
+            } => unreachable!(), // ???
             Layer::List {
                 remaining_elements: _,
                 element_tag: _,
             } => {
-                todo!();
+                unreachable!();
             }
         }
 
@@ -823,6 +870,15 @@ impl<'de, 'a> de::Deserializer<'de> for &'a mut Deserializer<'de> {
     }
 }
 
+fn try_size(size: i32, multiplier: usize) -> Result<usize> {
+    let size: usize = size
+        .try_into()
+        .map_err(|_| Error::bespoke("size was negative".to_string()))?;
+
+    size.checked_mul(multiplier)
+        .ok_or_else(|| Error::bespoke("size too large".to_string()))
+}
+
 struct CompoundAccess<'a, 'de> {
     de: &'a mut Deserializer<'de>,
 }
@@ -952,20 +1008,29 @@ impl<'a, 'de> de::VariantAccess<'de> for UnitVariantAccess<'a, 'de> {
     where
         T: serde::de::DeserializeSeed<'de>,
     {
-        todo!("unit variant: newtype variant")
+        Err(de::Error::invalid_type(
+            Unexpected::UnitVariant,
+            &"newtype variant",
+        ))
     }
 
     fn tuple_variant<V>(self, _len: usize, _visitor: V) -> Result<V::Value>
     where
         V: de::Visitor<'de>,
     {
-        todo!("unit variant: variant")
+        Err(de::Error::invalid_type(
+            Unexpected::TupleVariant,
+            &"tuple variant",
+        ))
     }
 
     fn struct_variant<V>(self, _fields: &'static [&'static str], _visitor: V) -> Result<V::Value>
     where
         V: de::Visitor<'de>,
     {
-        todo!("unit variant: struct variant")
+        Err(de::Error::invalid_type(
+            Unexpected::StructVariant,
+            &"struct variant",
+        ))
     }
 }
