@@ -23,8 +23,11 @@
 //! the original NBT. Without these types, it is not possible to tell if some
 //! data came from a NBT List or an NBT Array.
 //!
+//! A limitation of these array types is that they cannot be used with serde's
+//! untagged enums. If this is important to you please open an issue.
+//!
 //! Use these in your own data structures. They all implement
-//! [`Deref`][`std::ops::Deref`] for dereferencing into the underlying `Vec`.
+//! [`Deref`][`std::ops::Deref`] for dereferencing into a slice`.
 //!
 //! For versions that borrow their data, see [`borrow`]. For more information
 //! about deserialization see [`de`].
@@ -44,7 +47,7 @@
 //!
 //!# fn main(){
 //!     let buf: &[u8] = unimplemented!("get a buffer from somewhere");
-//!     let section: Section = fastnbt::de::from_bytes(buf).unwrap();
+//!     let section: Section = fastnbt::from_bytes(buf).unwrap();
 //!     let states = section.block_states.unwrap();
 //!
 //!     for long in states.iter() {
@@ -69,7 +72,7 @@
 //!```no_run
 //! use std::borrow::Cow;
 //! use fastnbt::error::Result;
-//! use fastnbt::{de::from_bytes, Value};
+//! use fastnbt::{from_bytes, Value};
 //! use flate2::read::GzDecoder;
 //! use serde::Deserialize;
 //! use std::io::Read;
@@ -120,28 +123,34 @@
 //! deserializing to Rust objects directly.
 //!
 
-use serde::{Deserialize, Serialize};
+use ser::{Serializer, State};
+use serde::{de as serde_de, Deserialize, Serialize};
 
 pub mod borrow;
 pub mod de;
 pub mod error;
+pub mod ser;
 pub mod stream;
 
 mod arrays;
+mod de_arrays;
 mod value;
 
 pub use arrays::*;
 pub use value::*;
 
-pub(crate) mod de_arrays;
-
 #[cfg(test)]
 mod test;
 
-use std::convert::{TryFrom, TryInto};
+use crate::{
+    de::Deserializer,
+    error::{Error, Result},
+};
+use std::{convert::TryFrom, fmt::Display, io::Write};
 
 /// An NBT tag. This does not carry the value or the name of the data.
 #[derive(Deserialize, Debug, PartialEq, Clone, Copy)]
+#[cfg_attr(feature = "arbitrary1", derive(arbitrary::Arbitrary))]
 #[repr(u8)]
 pub enum Tag {
     /// Represents the end of a Compound object.
@@ -172,17 +181,13 @@ pub enum Tag {
     LongArray = 12,
 }
 
-pub(crate) const BYTE_ARRAY_TAG: u8 = 7;
-pub(crate) const INT_ARRAY_TAG: u8 = 11;
-pub(crate) const LONG_ARRAY_TAG: u8 = 12;
-
 // Crates exist to generate this code for us, but would add to our compile
-// times, so we instead right it out manually, the tags will very rarely change
+// times, so we instead write it out manually, the tags will very rarely change
 // so isn't a massive burden, but saves a significant amount of compile time.
 impl TryFrom<u8> for Tag {
     type Error = ();
 
-    fn try_from(value: u8) -> Result<Self, ()> {
+    fn try_from(value: u8) -> std::result::Result<Self, ()> {
         use Tag::*;
         Ok(match value {
             0 => End,
@@ -223,31 +228,117 @@ impl From<Tag> for u8 {
     }
 }
 
-/// Compile time NBT tag type. Useful for forcing a custom type to have a field
-/// that must be a given tag. Used for the Array types.
-#[derive(Serialize, Clone, Copy, PartialEq)]
-pub(crate) struct CompTag<const N: u8>;
-
-impl<'de, const N: u8> Deserialize<'de> for CompTag<N> {
-    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
-    where
-        D: serde::Deserializer<'de>,
-    {
-        let tag: u8 = Deserialize::deserialize(deserializer)?;
-        if tag != N {
-            Err(serde::de::Error::custom("unexpected array type"))
-        } else {
-            Ok(Self)
-        }
-    }
-}
-
-impl<const N: u8> std::fmt::Debug for CompTag<N> {
+impl Display for Tag {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        let tag = <u8 as TryInto<Tag>>::try_into(N);
-        match tag {
-            Ok(tag) => tag.fmt(f),
-            Err(_) => write!(f, "InvalidTag({})", N),
+        let s = match self {
+            Tag::End => "end",
+            Tag::Byte => "byte",
+            Tag::Short => "short",
+            Tag::Int => "int",
+            Tag::Long => "long",
+            Tag::Float => "float",
+            Tag::Double => "double",
+            Tag::ByteArray => "byte-array",
+            Tag::String => "string",
+            Tag::List => "list",
+            Tag::Compound => "compound",
+            Tag::IntArray => "int-array",
+            Tag::LongArray => "long-array",
+        };
+        f.write_str(s)
+    }
+}
+
+pub fn to_bytes<T: Serialize>(v: &T) -> Result<Vec<u8>> {
+    let mut result = vec![];
+    let mut serializer = Serializer {
+        writer: &mut result,
+        state: State::Compound {
+            current_field: String::new(),
+        },
+    };
+    v.serialize(&mut serializer)?;
+    Ok(result)
+}
+
+pub fn to_writer<T: Serialize, W: Write>(writer: W, v: &T) -> Result<()> {
+    let mut serializer = Serializer {
+        writer,
+        state: State::Compound {
+            current_field: String::new(),
+        },
+    };
+    v.serialize(&mut serializer)?;
+    Ok(())
+}
+
+/// Deserialize into a `T` from some NBT data. See the [`de`] module for more
+/// information.
+///
+/// ```no_run
+/// # use fastnbt::Value;
+/// # use flate2::read::GzDecoder;
+/// # use std::io;
+/// # use std::io::Read;
+/// # use fastnbt::error::Result;
+/// # fn main() -> Result<()> {
+/// # let some_reader = io::stdin();
+/// let mut decoder = GzDecoder::new(some_reader);
+/// let mut buf = vec![];
+/// decoder.read_to_end(&mut buf).unwrap();
+///
+/// let val: Value = fastnbt::from_bytes(buf.as_slice())?;
+/// # Ok(())
+/// # }
+/// ```
+///
+/// [`de`]: ./index.html
+pub fn from_bytes<'a, T>(input: &'a [u8]) -> Result<T>
+where
+    T: serde_de::Deserialize<'a>,
+{
+    from_bytes_with_opts(input, Default::default())
+}
+
+pub struct DeOpts {
+    /// Maximum number of bytes a list or array can be.
+    max_seq_len: usize,
+}
+
+impl DeOpts {
+    pub fn new() -> Self {
+        Default::default()
+    }
+
+    pub fn max_seq_len(mut self, value: usize) -> Self {
+        self.max_seq_len = value;
+        self
+    }
+}
+
+impl Default for DeOpts {
+    fn default() -> Self {
+        Self {
+            max_seq_len: 100_000,
         }
     }
 }
+
+pub fn from_bytes_with_opts<'a, T>(input: &'a [u8], opts: DeOpts) -> Result<T>
+where
+    T: serde_de::Deserialize<'a>,
+{
+    const GZIP_MAGIC_BYTES: [u8; 2] = [0x1f, 0x8b];
+
+    // Provide freindly error for the common case of passing GZip data to
+    // `from_bytes`. This would be invalid starting data for NBT anyway.
+    if input.starts_with(&GZIP_MAGIC_BYTES) {
+        return Err(Error::bespoke(
+            "from_bytes expects raw NBT, but input appears to be gzipped".to_string(),
+        ));
+    }
+
+    let mut des = Deserializer::from_bytes(input, opts);
+    let t = T::deserialize(&mut des)?;
+    Ok(t)
+}
