@@ -1,9 +1,7 @@
-use crate::{de::from_bytes, Value};
+use super::resources::{CHUNK_RAW, CHUNK_RAW_WITH_ENTITIES};
+use crate::{from_bytes, Value};
 use serde::Deserialize;
 
-const CHUNK_RAW: &[u8] = include_bytes!("resources/chunk.nbt");
-const CHUNK_RAW_WITH_ENTITIES: &[u8] = include_bytes!("resources/chunk1.14.nbt");
-
 #[test]
 fn unit_variant_enum_for_chunk_status() {
     // From https://minecraft.gamepedia.com/Chunk_format
@@ -76,12 +74,16 @@ fn tile_entities() {
     enum KnownEntity {
         #[serde(rename = "minecraft:bat")]
         Bat {
+            #[allow(dead_code)]
             #[serde(rename = "BatFlags")]
             bat_flags: i8,
         },
 
         #[serde(rename = "minecraft:creeper")]
-        Creeper { ignited: i8 },
+        Creeper {
+            #[allow(dead_code)]
+            ignited: i8,
+        },
     }
 
     let chunk: Chunk = from_bytes(CHUNK_RAW_WITH_ENTITIES).unwrap();
